Require Import Coq.Arith.PeanoNat.
Require Import Coq.ZArith.ZArith.
Require Import Coq.Logic.FunctionalExtensionality.
Require Import Coq.micromega.Lia.
Require Import Coq.Strings.String. Open Scope string_scope.
Require Import Coq.Lists.List. Import ListNotations.
Set Default Goal Selector "!".

Module Nat.
  Lemma eqb_spec: forall x y: nat, BoolSpec (x = y) (x <> y) (Nat.eqb x y).
  Proof.
    intros. destruct (Nat.eqb x y) eqn: E; constructor.
    - apply Nat.eqb_eq. assumption.
    - apply Nat.eqb_neq. assumption.
  Qed.
End Nat.

Definition ident := string.

Inductive hunop := HUFst | HUSnd | HUNot.
Inductive lunop := LUNot.
Inductive hbinop := HBPlus | HBMinus | HBPair.
Inductive lbinop := LBPlus | LBMinus.

Inductive hval: Type :=
| HVNum(n: nat)
| HVPair(fst snd: hval).
Definition lval: Type := nat.

Section WithBinop.
  Context {unop: Type} {binop: Type}.
  Inductive cmd: Type :=
  | CUnop(x: ident)(op: unop)(arg1: ident)
  | CBinop(x: ident)(op: binop)(arg1 arg2: ident)
  | CInput(x: ident)
  | COutput(x: ident)
  | CLoad(x: ident)(addr: ident)(offset: nat)
  | CStore(addr: ident)(offset: nat)(v: ident)
  | CAlloc(x: ident)(n: nat)
  | CLit(x: ident)(v: nat)
  | CSet(x: ident)(y: ident)
  | CSeq(c1 c2: cmd)
  | CIf(cond: ident)(thn els: cmd)
  | CWhile(cond: ident)(body: cmd)
  | CSkip.

  Fixpoint vars(c: cmd): ident -> Prop :=
    match c with
    | CSkip => fun _ => False
    | CAlloc x _ | CLit x _ | CInput x | COutput x =>
      fun i => i = x
    | CLoad x y _ | CStore x _ y | CUnop x _ y | CSet x y =>
      fun i => i = x \/ i = y
    | CBinop x _ y z => fun i => i = x \/ i = y \/ i = z
    | CSeq c1 c2 => fun i => vars c1 i \/ vars c2 i
    | CIf cond c1 c2 => fun i => i = cond \/ vars c1 i \/ vars c2 i
    | CWhile cond body => fun i => i = cond \/ vars body i
    end.
End WithBinop.
Arguments cmd: clear implicits.
Definition hcmd := cmd hunop hbinop.
Definition lcmd := cmd lunop lbinop.

Ltac inv H := inversion H; clear H; subst.

Module locals.
  Section WithV.
    Context {V: Type}.
    Local Notation t := (ident -> option V).
    Definition put(m: t)(k: ident)(v: V): t := fun k' => if String.eqb k k' then Some v else m k'.
    Definition get(m: t)(k: ident) := m k.

    Lemma get_put_cases: forall k1 k2 v1 v2 m,
        get (put m k1 v1) k2 = Some v2 ->
        k1 = k2 /\ v1 = v2 \/ k1 <> k2 /\ get m k2 = Some v2.
    Proof. unfold get, put. intros. destruct (String.eqb_spec k1 k2); intuition congruence. Qed.

    Lemma get_put_same: forall k m v, get (put m k v) k = Some v.
    Proof. unfold get, put. intros. rewrite String.eqb_refl. reflexivity. Qed.

    Lemma get_put_diff: forall k1 k2 m v o,
        k1 <> k2 ->
        get m k2 = o ->
        get (put m k1 v) k2 = o.
    Proof. unfold get, put. intros. destruct (String.eqb_spec k1 k2); intuition congruence. Qed.
  End WithV.
End locals.
Definition locals(V: Type) := ident -> option V.
Definition hlocals := locals hval.
Definition llocals := locals lval.

Definition mem := nat -> option nat.
Module mem.
  Definition put(m: mem)(k: nat)(v: nat): mem := fun k' => if Nat.eqb k k' then Some v else m k'.
  Definition get(m: mem)(k: nat) := m k.
  Definition union(m1 m2: mem): mem :=
    fun k => match m2 k with
             | Some v => Some v
             | None => m1 k
             end.
  Definition diff(m1 m2: mem): mem :=
    fun k => match m2 k with
             | Some _ => None
             | None => m1 k
             end.
  Definition disjoint(m1 m2: mem): Prop := forall k v1 v2, get m1 k = Some v1 -> get m2 k = Some v2 -> False.
  Definition extends(m1 m2: mem): Prop := forall k v, get m2 k = Some v -> get m1 k = Some v.

  Lemma get_put_cases: forall k1 k2 v1 v2 m,
      get (put m k1 v1) k2 = Some v2 ->
      k1 = k2 /\ v1 = v2 \/ k1 <> k2 /\ get m k2 = Some v2. Search BoolSpec nat.
  Proof. unfold get, put. intros. destruct (Nat.eqb_spec k1 k2); intuition congruence. Qed.

  Lemma get_put_same: forall k m v, get (put m k v) k = Some v.
  Proof. unfold get, put. intros. rewrite Nat.eqb_refl. reflexivity. Qed.

  Lemma get_put_diff: forall k1 k2 m v o,
      k1 <> k2 ->
      get m k2 = o ->
      get (put m k1 v) k2 = o.
  Proof. unfold get, put. intros. destruct (Nat.eqb_spec k1 k2); intuition congruence. Qed.

  Lemma put_union_commute: forall m1 m2 k v,
      put (union m1 m2) k v = union m1 (put m2 k v).
  Proof.
    unfold put, union. intros. extensionality k'.
    destruct (Nat.eqb_spec k k'); reflexivity.
  Qed.

  Lemma get_extends: forall m1 m2 k v,
      extends m2 m1 ->
      get m1 k = Some v ->
      get m2 k = Some v.
  Proof. unfold extends. intros. eauto. Qed.

  Lemma extends_put_lr: forall m1 m2 k v,
      extends m2 m1 ->
      extends (put m2 k v) (put m1 k v).
  Proof.
    unfold extends. intros. eapply get_put_cases in H0.
    intuition subst.
    - apply get_put_same.
    - eauto using get_put_diff.
  Qed.

  Lemma extends_diff: forall m1 m2,
      extends m1 (diff m1 m2).
  Proof.
    unfold extends, diff, get. intros. destruct (m2 k). 1: discriminate. assumption.
  Qed.

  Lemma extends_refl: forall m, extends m m.
  Proof. unfold extends. intros *. exact id. Qed.

  Lemma get_undiff: forall m1 m2 k v,
      get (diff m1 m2) k = Some v ->
      get m1 k = Some v.
  Proof.
    intros. eapply get_extends. 2: eassumption. apply extends_diff.
  Qed.

  Lemma get_diff_Some: forall m1 m2 k v,
      get m1 k = Some v ->
      get m2 k = None ->
      get (diff m1 m2) k = Some v.
  Proof.
    unfold get, diff. intros. rewrite H0. assumption.
  Qed.

  Lemma get_diff_None: forall m1 m2 k,
      get m1 k = None \/ get m2 k <> None ->
      get (diff m1 m2) k = None.
  Proof.
    unfold get, diff. intros. destruct H; destruct (m2 k); congruence.
  Qed.

  Lemma diff_put_l_None: forall m1 m2 k v,
      get m2 k = None ->
      diff (put m1 k v) m2 = put (diff m1 m2) k v.
  Proof.
    unfold get, diff, put. intros. extensionality k'. destruct (Nat.eqb_spec k k').
    - subst k'. rewrite H. reflexivity.
    - reflexivity.
  Qed.

  Lemma diff_put_l_Some: forall m1 m2 k v1 v2,
      get m2 k = Some v2 ->
      diff (put m1 k v1) m2 = diff m1 m2.
  Proof.
    unfold diff, put, get. intros. extensionality k'. destruct (m2 k') eqn: E. 1: reflexivity.
    destruct (Nat.eqb_spec k k'); congruence.
  Qed.

  Lemma diff_put_r_None: forall m1 m2 k v,
      get m1 k = None ->
      diff m1 (put m2 k v) = diff m1 m2.
  Proof.
    unfold get, diff, put. intros. extensionality k'. destruct (Nat.eqb_spec k k'). 2: reflexivity.
    subst k'. rewrite H. destruct (m2 k); reflexivity.
  Qed.

  Lemma diff_put_r_Some: forall m1 m2 k v v_old,
      get m2 k = Some v_old ->
      diff m1 (put m2 k v) = diff m1 m2.
  Proof.
    unfold get, diff, put. intros. extensionality k'. destruct (Nat.eqb_spec k k'). 2: reflexivity.
    subst k'. rewrite H. reflexivity.
  Qed.

  Lemma extends_preserves_None: forall m1 m2 k,
      extends m2 m1 ->
      get m2 k = None ->
      get m1 k = None.
  Proof.
    unfold extends. intros. destruct (get m1 k) eqn: E; [exfalso|reflexivity].
    specialize H with (1 := E). congruence.
  Qed.

  Lemma extends_put_l: forall m1 m2 k v,
      get m1 k = None ->
      extends m2 m1 ->
      extends (put m2 k v) m1.
  Proof.
    unfold extends. intros. specialize H0 with (1 := H1).
    unfold get, put in *.
    destruct (Nat.eqb_spec k k0); congruence.
  Qed.

  Lemma put_idemp: forall m k v,
      get m k = Some v ->
      put m k v = m.
  Proof.
    unfold get, put. intros. extensionality k'. destruct (Nat.eqb_spec k k'); congruence.
  Qed.

  Lemma put_eq_union: forall m1 m2 m3 k v,
      m1 = union m2 m3 ->
      put m1 k v = union m2 (put m3 k v).
  Proof. intros. subst. apply put_union_commute. Qed.

  Lemma invert_get_union_None: forall m1 m2 k,
      get (union m1 m2) k = None ->
      get m1 k = None /\ get m2 k = None.
  Proof.
    unfold get, union. intros. destruct (m2 k). 1: discriminate. auto.
  Qed.

  Lemma put_put_same: forall m k v1 v2,
      put (put m k v1) k v2 = put m k v2.
  Proof. unfold get, put. intros. extensionality x. destruct (Nat.eqb_spec k x); reflexivity. Qed.

  Lemma get_disjoint_union_l: forall m1 m2 k v,
      disjoint m1 m2 ->
      get m1 k = Some v ->
      get (union m1 m2) k = Some v.
  Proof.
    unfold disjoint, get, union. intros. subst. destruct (m2 k) eqn: E.
    - exfalso. eauto.
    - assumption.
  Qed.

  Lemma get_union_r: forall m1 m2 k v,
      get m2 k = Some v ->
      get (union m1 m2) k = Some v.
  Proof.
    unfold get, union. intros. rewrite H. reflexivity.
  Qed.

  Definition unallocated(m: mem): nat -> nat -> Prop :=
    fix rec addr count :=
      match count with
      | O => True
      | S count' => get m addr = None /\ rec (S addr) count'
      end.
  Fixpoint put_range(m: mem)(a: nat)(vs: list nat): mem :=
    match vs with
    | nil => m
    | cons h t => put (put_range m (S a) t) a h
    end.

  Lemma extends_preserves_unallocated: forall m1 m2 n a,
      extends m2 m1 ->
      unallocated m2 a n ->
      unallocated m1 a n.
  Proof.
    intros. revert a H0. induction n; intros; simpl in *; intuition eauto using extends_preserves_None.
  Qed.

  Lemma extends_put_range: forall m1 m2 vs a,
      extends m2 m1 ->
      extends (put_range m2 a vs) (put_range m1 a vs).
  Proof.
    induction vs; intros; simpl; eauto using extends_put_lr.
  Qed.

  Lemma get_put_range_outside: forall m vs a1 a2,
      ~ (a1 <= a2 < a1 + List.length vs) ->
      get (put_range m a1 vs) a2 = get m a2.
  Proof.
    induction vs; simpl; intros.
    - reflexivity.
    - eapply get_put_diff. 1: lia. eapply IHvs. lia.
  Qed.

  Lemma diff_put_range_lr: forall m1 m2 vs a,
      unallocated m1 a (List.length vs) ->
      diff (put_range m1 a vs) (put_range m2 a vs) = diff m1 m2.
  Proof.
    induction vs; simpl; intros. 1: reflexivity. destruct H.
    erewrite diff_put_l_Some by eapply get_put_same.
    erewrite diff_put_r_None. 2: {
      rewrite get_put_range_outside by lia. assumption.
    }
    eapply IHvs.
    assumption.
  Qed.

End mem.

Inductive hevalunop: hunop -> hval -> hval -> Prop :=
| HEFst: forall v1 v2,
    hevalunop HUFst (HVPair v1 v2) v1
| HESnd: forall v1 v2,
    hevalunop HUSnd (HVPair v1 v2) v2
| HENotZero:
    hevalunop HUNot (HVNum 0) (HVNum 1)
| HENotOne:
    hevalunop HUNot (HVNum 1) (HVNum 0).

Inductive hevalbinop: hbinop -> hval -> hval -> hval -> Prop :=
| HEPlus: forall n1 n2,
    hevalbinop HBPlus (HVNum n1) (HVNum n2) (HVNum (n1 + n2))
| HEMinus: forall n1 n2,
    hevalbinop HBMinus (HVNum n1) (HVNum n2) (HVNum (n1 - n2))
| HEPair: forall v1 v2,
    hevalbinop HBPair v1 v2 (HVPair v1 v2).

Inductive levalbinop: lbinop -> lval -> lval -> lval -> Prop :=
| LEPlus: forall n1 n2,
    levalbinop LBPlus n1 n2 (n1 + n2)
| LEMinus: forall n1 n2,
    levalbinop LBMinus n1 n2 (n1 - n2).

Inductive levalunop: lunop -> lval -> lval -> Prop :=
| LENotZero: levalunop LUNot 0 1
| LENotOne: levalunop LUNot 1 0.

Inductive Event :=
| IN(x: nat)
| OUT(x: nat).

Definition trace := list Event.

(* Naming:
   eval[ao][nd][tp][bs] for all/one, nondet/det, total/partial, big/small
   - all = one derivation talks about all executions (omnisemantics)
   - one = one derivation talks about one possible execution (traditional style)
   Note that omnisemantics for deterministic semantics is useful too,
   because instead of writing "exists s', s --> s' /\ P s'", we can just write
   "s --> P", which simplifies proving such judgments.
   Therefore, we keep the all/one dimension orthogonal to the nondet/det
   dimension, rather than collapsing it into all/one/det.
 *)

Implicit Type t: trace.
Implicit Type m: mem.

Section GenericSemantics.
  Context {V: Type}
          {unop: Type} {evalunop: unop -> V -> V -> Prop}
          {binop: Type} {evalbinop: binop -> V -> V -> V -> Prop}
          {nat2val: nat -> V}.

  Implicit Type post: trace -> mem -> (ident -> option V) -> Prop.
  Implicit Type l: locals V.

  (* omni-bigstep semantics *)
  Inductive evalantb: cmd unop binop -> trace -> mem -> locals V ->
                      (trace -> mem -> locals V -> Prop) -> Prop :=
  | evalantb_unop: forall x op y vy r t m l post,
      locals.get l y = Some vy ->
      evalunop op vy r ->
      post t m (locals.put l x r) ->
      evalantb (CUnop x op y) t m l post
  | evalantb_binop: forall x op y z vy vz r t m l post,
      locals.get l y = Some vy ->
      locals.get l z = Some vz ->
      evalbinop op vy vz r ->
      post t m (locals.put l x r) ->
      evalantb (CBinop x op y z) t m l post
  | evalantb_input: forall x t m l post,
      (forall n, post (IN n :: t) m (locals.put l x (nat2val n))) ->
      evalantb (CInput x) t m l post
  | evalantb_output: forall x t m l n post,
      locals.get l x = Some (nat2val n) ->
      post (OUT n :: t) m l ->
      evalantb (COutput x) t m l post
  | evalantb_load: forall x a addr offset t m l post v,
      locals.get l a = Some (nat2val addr) ->
      mem.get m (offset + addr) = Some v ->
      post t m (locals.put l x (nat2val v)) ->
      evalantb (CLoad x a offset) t m l post
  | evalantb_store: forall x y offset v t m l a old post,
      locals.get l x = Some (nat2val a) ->
      locals.get l y = Some (nat2val v) ->
      mem.get m (offset + a) = Some old ->
      post t (mem.put m (offset + a) v) l ->
      evalantb (CStore x offset y) t m l post
  | evalantb_alloc: forall x n t m l post,
      (* non-deterministically pick an address a at which n
         nondeterministically chosen initial values can be stored *)
      (forall a vs,
          List.length vs = n ->
          mem.unallocated m a n ->
          post t (mem.put_range m a vs) (locals.put l x (nat2val a))) ->
      evalantb (CAlloc x n) t m l post
  | evalantb_lit: forall x v t m l post,
      post t m (locals.put l x (nat2val v)) ->
      evalantb (CLit x v) t m l post
  | evalantb_set: forall x y v t m l post,
      locals.get l y = Some v ->
      post t m (locals.put l x v) ->
      evalantb (CSet x y) t m l post
  | evalantb_seq: forall c1 c2 t m l mid post,
      evalantb c1 t m l mid ->
      (forall t' m' l', mid t' m' l' -> evalantb c2 t' m' l' post) ->
      evalantb (CSeq c1 c2) t m l post
  | evalantb_ifTrue: forall cond thn els t m l post,
      locals.get l cond = Some (nat2val 1) ->
      evalantb thn t m l post ->
      evalantb (CIf cond thn els) t m l post
  | evalantb_ifFalse: forall cond thn els t m l post,
      locals.get l cond = Some (nat2val 0) ->
      evalantb els t m l post ->
      evalantb (CIf cond thn els) t m l post
  | evalantb_whileDone: forall cond body t m l post,
      locals.get l cond = Some (nat2val 0) ->
      post t m l ->
      evalantb (CWhile cond body) t m l post
  | evalantb_whileAgain: forall cond body t m l mid post,
      locals.get l cond = Some (nat2val 1) ->
      evalantb body t m l mid ->
      (forall t' m' l', mid t' m' l' ->
                        evalantb (CWhile cond body) t' m' l' post) ->
      evalantb (CWhile cond body) t m l post
  | evalantb_skip: forall t m l post,
      post t m l ->
      evalantb CSkip t m l post.

  Hint Constructors evalantb : evalantb.

  Lemma evalantb_seq_chained: forall c1 c2 t m l post,
      evalantb c1 t m l (fun t' m' l' => evalantb c2 t' m' l' post) ->
      evalantb (CSeq c1 c2) t m l post.
  Proof. eauto with evalantb. Qed.
End GenericSemantics.

Definition hevalantb: hcmd -> trace -> mem -> hlocals -> (trace -> mem -> hlocals -> Prop) -> Prop :=
  @evalantb hval hunop hevalunop hbinop hevalbinop HVNum.
Definition levalantb: lcmd -> trace -> mem -> llocals -> (trace -> mem -> llocals -> Prop) -> Prop :=
  @evalantb lval lunop levalunop lbinop levalbinop id.

Notation "c1 ;; c2" := (CSeq c1 c2) (at level 51, right associativity).

Definition compile_unop(x: ident)(op: hunop)(y: ident): lcmd :=
  match op with
  | HUFst => CLoad x y 0
  | HUSnd => CLoad x y 1
  | HUNot => CUnop x LUNot y
  end.

Definition compile_binop(x: ident)(op: hbinop)(y z: ident): lcmd :=
  match op with
  | HBPlus => CBinop x LBPlus y z
  | HBMinus => CBinop x LBMinus y z
  (* "_tmp" is a reserved name. We can't use x directly, because if x=y or x=z, we would overwrite it *)
  | HBPair => CAlloc "_tmp" 2;; CStore "_tmp" 0 y;; CStore "_tmp" 1 z;; CSet x "_tmp"
  end.

Fixpoint compile(c: hcmd): lcmd :=
  match c with
  | CUnop x op y => compile_unop x op y
  | CBinop x op y z => compile_binop x op y z
  | CInput x => CInput x
  | COutput x => COutput x
  | CLoad x y offset => CLoad x y offset
  | CStore x offset y => CStore x offset y
  | CAlloc x n => CAlloc x n
  | CLit x v => CLit x v
  | CSet x y => CSet x y
  | CSeq c1 c2 => CSeq (compile c1) (compile c2)
  | CIf cond thn els => CIf cond (compile thn) (compile els)
  | CWhile cond body => CWhile cond (compile body)
  | CSkip => CSkip
  end.

Notation len := List.length.
Coercion Z.of_nat : nat >-> Z.

(* high-level value v is represented by low-level value w *)
Fixpoint valuerepr(v: hval)(w: nat)(m: mem): Prop :=
  match v with
  | HVNum n => w = n
  | HVPair v1 v2 => (exists w1, mem.get m w = Some w1 /\ valuerepr v1 w1 m) /\
                    (exists w2, mem.get m (S w) = Some w2 /\ valuerepr v2 w2 m)
  end.

(* mH: high-level memory
   mL: low-level memory *)
Definition R(mH: mem)(lH: hlocals)(mL: mem)(lL: llocals): Prop :=
  locals.get lH "_tmp" = None /\
  mem.extends mL mH /\
  forall x v, locals.get lH x = Some v ->
              exists w, locals.get lL x = Some w /\ valuerepr v w (mem.diff mL mH).

#[local] Hint Constructors evalantb : core.

(* `Known P` remains in the goal even if `H: P` is destructed, which avoids
   infinitite specialization loops *)
Inductive Known(P: Prop): Prop := mkKnown.

Ltac record p :=
  let P := type of p in
  lazymatch goal with
  | H: Known P |- _ => fail 0 P "is already known"
  | |- _ => let H := fresh "K0" in pose proof mkKnown P as H; move H at top
  end.

Lemma destruct_length_S{A: Type}: forall (l: list A) n,
    List.length l = S n ->
    exists h rest, l = cons h rest /\ List.length rest = n.
Proof. intros. destruct l. 1: discriminate. simpl in *. inv H. eauto. Qed.

Lemma destruct_length_O{A: Type}: forall (l: list A),
    List.length l = O ->
    l = nil.
Proof. intros. destruct l. 2: discriminate. reflexivity. Qed.

Lemma valuerepr_put: forall v1 v2 a1 a2 m,
    mem.get m a2 = None ->
    valuerepr v1 a1 m ->
    valuerepr v1 a1 (mem.put m a2 v2).
Proof.
  induction v1; intros; cbn in *.
  - assumption.
  - destruct H0 as ((w1 & ? & ?) & (w2 & ? & ?)).
    specialize IHv1_1 with (1 := H) (2 := H1).
    specialize IHv1_2 with (1 := H) (2 := H3).
    split.
    + eexists. split.
      * apply mem.get_put_diff. 1: congruence. eassumption.
      * eapply IHv1_1.
    + eexists. split.
      * apply mem.get_put_diff. 1: congruence. eassumption.
      * eapply IHv1_2.
Qed.

Lemma valuerepr_extends: forall m1 m2 v a,
    mem.extends m2 m1 ->
    valuerepr v a m1 ->
    valuerepr v a m2.
Proof.
  intros. revert v a H0. induction v; intros; cbn in *.
  - assumption.
  - destruct H0 as ((w1 & ? & ?) & (w2 & ? & ?)).
    specialize IHv1 with (1 := H1).
    specialize IHv2 with (1 := H3).
    split.
    + eexists. split. 2: eassumption.
      unfold mem.extends in *. eauto.
    + eexists. split. 2: eassumption.
      unfold mem.extends in *. eauto.
Qed.

Lemma mem_put_dedup1: forall m k1 k2 v1a v1b v2a v2b,
    mem.put (mem.put (mem.put (mem.put m k1 v1a) k2 v2a) k2 v2b) k1 v1b =
    mem.put (mem.put m k2 v2b) k1 v1b.
Proof.
  unfold mem.put. intros. extensionality k'.
  destruct (Nat.eqb_spec k1 k'); destruct (Nat.eqb_spec k2 k'); reflexivity.
Qed.

Inductive CurrentConstructor{A: Type}(c: A): Prop := mkCurrentConstructor.

Ltac evalantb_induct H :=
  induction H;
  [ let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_unop) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_binop) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_input) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_output) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_load) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_store) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_alloc) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_lit) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_set) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_seq) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_ifTrue) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_ifFalse) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_whileDone) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_whileAgain) as C; move C at top
  | let C := fresh "C" in pose proof mkCurrentConstructor (@evalantb_skip) as C; move C at top ].

Ltac head_of_app e :=
  match e with
  | ?f ?a => head_of_app f
  | _ => e
  end.

Ltac step :=
  match goal with
  | H: exists x, _ |- _ => let y := fresh x in destruct H as (y & H)
  | H: _ /\ _ |- _ => let L := fresh H "_0" in destruct H as (L & H)
  | |- forall _, _ => intros
  | H: List.length _ = S _ |- _ => eapply destruct_length_S in H
  | H: List.length _ = O |- _ => eapply destruct_length_O in H
  | H: mem.get (mem.union _ _) _ = _ |- _ => eapply mem.invert_get_union_None in H
  | H: hevalunop _ _ _ |- _ => inv H
  | H: hevalbinop _ _ _ _ |- _ => inv H
  | H1: ?P, H2: _ |- _ =>
    lazymatch type of P with
    | Prop => let F := fresh H1 "_" H2 in specialize H1 with (1 := H2) as F; record F
    end
  | |- _ => progress (unfold id, levalantb; simpl)
  | |- _ => progress subst
  | |- context [mem.put (mem.put (mem.put (mem.put ?m ?k1 ?v1a) ?k2 ?v2a) ?k2 ?v2b) ?k1 ?v1b] =>
    rewrite (mem_put_dedup1 m k1 k2 v1a v1b v2a v2b)
  | H: ?P |- _ => lazymatch P with
                  | Known _ => fail
                  | _ => progress simpl in H
                  end
  | |- _ /\ _ => split
  | |- ?post _ _ _ =>
    let t := type of post in
    let t' := open_constr:(trace -> mem -> locals _ -> Prop) in
    unify t t'; eassumption
  | H: forall (n: nat), ?post _ _ _ |- ?post _ _ _ => apply H
  | |- ?G => assert_fails (has_evar G);
             first [ solve [eauto using mem.extends_preserves_None, mem.extends_refl, mem.extends_put_range]
                   | eapply mem.extends_put_l;
                     [ solve[eauto using mem.extends_preserves_None, mem.get_diff_None] | ]
                   | eapply mem.extends_put_lr
                   | erewrite mem.diff_put_l_Some by eauto
                   | erewrite mem.diff_put_r_Some by eauto
                   | rewrite mem.diff_put_range_lr by eauto
                   | rewrite mem.diff_put_l_None by eauto using mem.extends_preserves_None ]
  | |- ?LHS = ?RHS => assert_fails (has_evar LHS);
                      first [ solve [eauto using mem.get_disjoint_union_l, mem.get_union_r, mem.get_undiff]
                            | eapply mem.get_diff_Some
                            | eapply mem.get_diff_None ]
  | |- ?LHS = ?RHS => assert_fails (has_evar LHS);
                      first [ eassumption
                            | apply locals.get_put_same
                            | eapply locals.get_put_diff; [intuition congruence|]]
  | |- ?LHS = ?RHS => assert_fails (has_evar LHS);
                      first [ eassumption
                            | apply mem.get_put_same
                            | eapply mem.get_put_diff; [lia || congruence |]]
  | |- _ => eapply valuerepr_extends; [|eassumption]
  | |- _ => eapply LENotZero
  | |- _ => eapply LENotOne
  | |- _ => eapply LEPlus
  | |- _ => eapply LEMinus
  | |- _ => eapply evalantb_unop
  | |- _ => eapply evalantb_binop
  | |- _ => eapply evalantb_input
  | |- _ => eapply evalantb_output
  | |- _ => eapply evalantb_load
  | |- _ => eapply evalantb_store
  | |- _ => eapply evalantb_alloc
  | |- _ => eapply evalantb_lit
  | |- _ => eapply evalantb_set
  | |- evalantb (CSeq (compile ?c1) _) _ _ _ _ => is_var c1; eapply evalantb_seq
  | |- evalantb (CSeq ?c1 _) _ _ _ _ => let h := head_of_app c1 in is_constructor h; eapply evalantb_seq_chained
  | |- _ => eapply evalantb_ifTrue; [solve[eauto]|]
  | |- _ => eapply evalantb_ifFalse; [solve[eauto]|]
  | |- _ => eapply evalantb_whileDone; [solve[eauto]|]
  | |- _ => eapply evalantb_whileAgain; [solve[eauto]| | ]
  | |- _ => eapply evalantb_skip
  | |- _ => eapply evalantb_seq_chained
  | H: _ |- _ => eapply locals.get_put_cases in H
  | H: _ |- _ => eapply mem.get_put_cases in H
  | H1: mem.extends _ _, H2: mem.unallocated _ _ _ |- _ =>
    let U := fresh H1 H2 in pose proof (mem.extends_preserves_unallocated _ _ _ _ H1 H2) as U; record U
  | H: _ \/ _ |- _ => destruct H as [H | H]
  | |- exists _, _ => eexists
  | IH: _ |- evalantb (compile ?c) _ _ _ _ => is_var c; eapply IH
  | IH: _ |- evalantb (CWhile _ (compile _)) _ _ _ _ => eapply IH
  end.

Set Default Goal Selector "1".

Lemma compile_correct: forall c t mH lH mL lL post,
    ~ vars c "_tmp" ->
    R mH lH mL lL ->
    hevalantb c t mH lH post ->
    levalantb (compile c) t mL lL (fun t' mL' lL' =>
       exists mH' lH', post t' mH' lH' /\ R mH' lH' mL' lL').
Proof.
  intros *. intros HV HR HE. revert mL lL HV HR. unfold R.
   evalantb_induct HE; try solve [repeat step]. (* takes about 30s *)
Qed.

Print Assumptions compile_correct.
